@@ -248,7 +248,7 @@ soap_sax_characters (gpointer _ctxt,
 			g_warning ("Failed to write streaming data to file");
 		}
 	} else {
-		guchar *bdata = g_malloc (len);
+		guchar *bdata = g_malloc ((len * 3 / 4) + 3);
 		gsize blen;
 
 		blen = g_base64_decode_step (
@@ -1135,7 +1135,7 @@ e_soap_message_get_xml_doc (ESoapMessage *msg)
  *
  * Parses the response returned by the server.
  *
- * Returns: a #ESoapResponse representing the response from
+ * Returns: an #ESoapResponse representing the response from
  * the server, or %NULL if there was an error.
  */
 ESoapResponse *
