@@ -17,15 +17,12 @@
 #include <mail/e-mail-config-receiving-page.h>
 
 #include "common/camel-ews-settings.h"
+#include "common/e-oauth2-service-office365.h"
 
 #include "e-mail-config-ews-autodiscover.h"
 #include "e-ews-config-utils.h"
 #include "e-ews-search-user.h"
 
-#define E_MAIL_CONFIG_EWS_BACKEND_GET_PRIVATE(obj) \
-	(G_TYPE_INSTANCE_GET_PRIVATE \
-	((obj), E_TYPE_MAIL_CONFIG_EWS_BACKEND, EMailConfigEwsBackendPrivate))
-
 struct _EMailConfigEwsBackendPrivate {
 	GtkWidget *user_entry;
 	GtkWidget *host_entry;
@@ -38,12 +35,12 @@ struct _EMailConfigEwsBackendPrivate {
 	GtkWidget *oauth2_tenant_entry;
 	GtkWidget *oauth2_client_id_entry;
 	GtkWidget *oauth2_redirect_uri_entry;
+	GtkWidget *oauth2_resource_uri_entry;
+	GtkWidget *oauth2_endpoint_host_entry;
 };
 
-G_DEFINE_DYNAMIC_TYPE (
-	EMailConfigEwsBackend,
-	e_mail_config_ews_backend,
-	E_TYPE_MAIL_CONFIG_SERVICE_BACKEND)
+G_DEFINE_DYNAMIC_TYPE_EXTENDED (EMailConfigEwsBackend, e_mail_config_ews_backend, E_TYPE_MAIL_CONFIG_SERVICE_BACKEND, 0,
+	G_ADD_PRIVATE_DYNAMIC (EMailConfigEwsBackend))
 
 static ESource *
 mail_config_ews_backend_new_collection (EMailConfigServiceBackend *backend)
@@ -76,7 +73,7 @@ static void
 search_for_impersonate_user_clicked_cb (GtkButton *button,
 					EMailConfigServiceBackend *backend)
 {
-	EMailConfigEwsBackendPrivate *priv;
+	EMailConfigEwsBackend *ews_backend;
 	ESource *source;
 	CamelSettings *settings;
 	EEwsConnection *conn;
@@ -89,13 +86,13 @@ search_for_impersonate_user_clicked_cb (GtkButton *button,
 	if (!source)
 		source = e_mail_config_service_backend_get_source (backend);
 
-	priv = E_MAIL_CONFIG_EWS_BACKEND_GET_PRIVATE (backend);
+	ews_backend = E_MAIL_CONFIG_EWS_BACKEND (backend);
 	settings = e_mail_config_service_backend_get_settings (backend);
-	conn = e_ews_connection_new (source, gtk_entry_get_text (GTK_ENTRY (priv->host_entry)), CAMEL_EWS_SETTINGS (settings));
+	conn = e_ews_connection_new (source, gtk_entry_get_text (GTK_ENTRY (ews_backend->priv->host_entry)), CAMEL_EWS_SETTINGS (settings));
 	parent = e_ews_config_utils_get_widget_toplevel_window (GTK_WIDGET (button));
 
 	if (e_ews_search_user_modal (parent, conn, NULL, NULL, &email)) {
-		gtk_entry_set_text (GTK_ENTRY (priv->impersonate_user_entry), email);
+		gtk_entry_set_text (GTK_ENTRY (ews_backend->priv->impersonate_user_entry), email);
 	}
 
 	g_object_unref (conn);
@@ -136,19 +133,25 @@ static void
 mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
                                         GtkBox *parent)
 {
-	EMailConfigEwsBackendPrivate *priv;
+	EMailConfigEwsBackend *ews_backend;
 	EMailConfigServicePage *page;
 	ESource *source;
 	ESourceExtension *extension;
 	CamelSettings *settings;
+	CamelEwsSettings *ews_settings;
 	GtkLabel *label;
 	GtkWidget *widget;
 	GtkWidget *container;
+	GtkWidget *expander;
+	GtkWidget *advanced_help;
+	GtkWidget *endpoint_host_label;
+	GtkWidget *redirect_uri_label;
+	GtkWidget *resource_uri_label;
 	const gchar *extension_name;
 	const gchar *text;
 	gchar *markup;
 
-	priv = E_MAIL_CONFIG_EWS_BACKEND_GET_PRIVATE (backend);
+	ews_backend = E_MAIL_CONFIG_EWS_BACKEND (backend);
 	page = e_mail_config_service_backend_get_page (backend);
 
 	/* This backend serves double duty.  One instance holds the
@@ -192,7 +195,7 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_label_set_mnemonic_widget (label, widget);
 	gtk_grid_attach (GTK_GRID (container), widget, 1, 0, 2, 1);
-	priv->user_entry = widget;  /* do not reference */
+	ews_backend->priv->user_entry = widget;  /* do not reference */
 	gtk_widget_show (widget);
 
 	widget = gtk_label_new_with_mnemonic (_("_Host URL:"));
@@ -206,12 +209,12 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_label_set_mnemonic_widget (label, widget);
 	gtk_grid_attach (GTK_GRID (container), widget, 1, 1, 1, 1);
-	priv->host_entry = widget;  /* do not reference */
+	ews_backend->priv->host_entry = widget;  /* do not reference */
 	gtk_widget_show (widget);
 
 	widget = e_mail_config_ews_autodiscover_new (backend);
 	gtk_grid_attach (GTK_GRID (container), widget, 2, 1, 1, 1);
-	priv->url_button = widget;  /* do not reference */
+	ews_backend->priv->url_button = widget;  /* do not reference */
 	gtk_widget_show (widget);
 
 	widget = gtk_label_new_with_mnemonic (_("OAB U_RL:"));
@@ -225,22 +228,32 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_label_set_mnemonic_widget (label, widget);
 	gtk_grid_attach (GTK_GRID (container), widget, 1, 2, 2, 1);
-	priv->oab_entry = widget;  /* do not reference */
+	ews_backend->priv->oab_entry = widget;  /* do not reference */
 	gtk_widget_show (widget);
 
 	widget = gtk_check_button_new_with_mnemonic (_("Open _Mailbox of other user"));
 	gtk_grid_attach (GTK_GRID (container), widget, 1, 3, 1, 1);
 	gtk_widget_show (widget);
 
-	if (camel_ews_settings_get_use_impersonation (CAMEL_EWS_SETTINGS (settings))) {
-		const gchar *impersonate_user = camel_ews_settings_get_impersonate_user (CAMEL_EWS_SETTINGS (settings));
+	ews_settings = CAMEL_EWS_SETTINGS (settings);
+
+	camel_ews_settings_lock (ews_settings);
+
+	if (camel_ews_settings_get_use_impersonation (ews_settings)) {
+		const gchar *impersonate_user = camel_ews_settings_get_impersonate_user (ews_settings);
 
 		if (impersonate_user && !*impersonate_user) {
-			camel_ews_settings_set_impersonate_user (CAMEL_EWS_SETTINGS (settings), NULL);
-			camel_ews_settings_set_use_impersonation (CAMEL_EWS_SETTINGS (settings), FALSE);
+			camel_ews_settings_unlock (ews_settings);
+
+			camel_ews_settings_set_impersonate_user (ews_settings, NULL);
+			camel_ews_settings_set_use_impersonation (ews_settings, FALSE);
+
+			camel_ews_settings_lock (ews_settings);
 		}
 	}
 
+	camel_ews_settings_unlock (ews_settings);
+
 	e_binding_bind_property (
 		settings, "use-impersonation",
 		widget, "active",
@@ -250,7 +263,7 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_grid_attach (GTK_GRID (container), widget, 1, 4, 1, 1);
 	gtk_widget_show (widget);
-	priv->impersonate_user_entry = widget;  /* do not reference */
+	ews_backend->priv->impersonate_user_entry = widget;  /* do not reference */
 
 	e_binding_bind_object_text_property (
 		settings, "impersonate-user",
@@ -267,7 +280,7 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_widget_show (widget);
 
 	e_binding_bind_property (
-		priv->impersonate_user_entry, "sensitive",
+		ews_backend->priv->impersonate_user_entry, "sensitive",
 		widget, "sensitive",
 		G_BINDING_SYNC_CREATE);
 
@@ -286,23 +299,23 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	widget = e_mail_config_auth_check_new (backend);
 	gtk_widget_set_margin_left (widget, 12);
 	gtk_box_pack_start (GTK_BOX (parent), widget, FALSE, FALSE, 0);
-	priv->auth_check = widget;  /* do not reference */
+	ews_backend->priv->auth_check = widget;  /* do not reference */
 	gtk_widget_show (widget);
 
 	widget = gtk_grid_new ();
 	gtk_widget_set_margin_left (widget, 12);
 	gtk_box_pack_start (GTK_BOX (parent), widget, FALSE, FALSE, 0);
-	priv->oauth2_settings_grid = GTK_GRID (widget);
+	ews_backend->priv->oauth2_settings_grid = GTK_GRID (widget);
 
-	gtk_grid_set_column_spacing (priv->oauth2_settings_grid, 4);
-	gtk_grid_set_row_spacing (priv->oauth2_settings_grid, 4);
+	gtk_grid_set_column_spacing (ews_backend->priv->oauth2_settings_grid, 4);
+	gtk_grid_set_row_spacing (ews_backend->priv->oauth2_settings_grid, 4);
 
 	container = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
-	gtk_grid_attach (priv->oauth2_settings_grid, container, 0, 0, 2, 1);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, container, 0, 0, 2, 1);
 
 	widget = gtk_check_button_new_with_mnemonic (_("_Override Office365 OAuth2 settings"));
 	gtk_box_pack_start (GTK_BOX (container), widget, FALSE, FALSE, 0);
-	priv->oauth2_override_check = widget;
+	ews_backend->priv->oauth2_override_check = widget;
 
 	markup = g_markup_printf_escaped ("(<a href=\"https://wiki.gnome.org/Apps/Evolution/EWS/OAuth2\">%s</a>)", _("Help…"));
 	widget = gtk_label_new (markup);
@@ -311,57 +324,25 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	gtk_box_pack_start (GTK_BOX (container), widget, FALSE, FALSE, 0);
 	g_free (markup);
 
-	/* Translators: 'Tenant' here means a term used by Microsoft to identify a company or organization in an Office 365 world.
-	   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
-	widget = gtk_label_new_with_mnemonic (_("_Tenant:"));
-	gtk_widget_set_margin_left (widget, 12);
-	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 0, 1, 1, 1);
-	label = GTK_LABEL (widget);
-
-	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
-		widget, "sensitive",
-		G_BINDING_SYNC_CREATE);
-
-	widget = gtk_entry_new ();
-	gtk_widget_set_hexpand (widget, TRUE);
-	gtk_label_set_mnemonic_widget (label, widget);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 1, 1, 1, 1);
-	priv->oauth2_tenant_entry = widget;
-
-	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
-		widget, "sensitive",
-		G_BINDING_SYNC_CREATE);
-
-	mail_config_ews_backend_set_oauth2_tooltip (widget, OFFICE365_TENANT,
-		/* Translators: 'Tenant' here means a term used by Microsoft to identify a company or organization in an Office 365 world. Same for 'common', it's a default URL path.
-		   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
-		_("Default tenant is “common“"),
-		/* Translators: 'Tenant' here means a term used by Microsoft to identify a company or organization in an Office 365 world.
-		   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
-		g_strdup_printf (_("Default tenant is “%s”"), OFFICE365_TENANT));
-
 	widget = gtk_label_new_with_mnemonic (_("Application I_D:"));
 	gtk_widget_set_margin_left (widget, 12);
 	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 0, 2, 1, 1);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 0, 1, 1, 1);
 	label = GTK_LABEL (widget);
 
 	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
+		ews_backend->priv->oauth2_override_check, "active",
 		widget, "sensitive",
 		G_BINDING_SYNC_CREATE);
 
 	widget = gtk_entry_new ();
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_label_set_mnemonic_widget (label, widget);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 1, 2, 1, 1);
-	priv->oauth2_client_id_entry = widget;
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 1, 1, 1, 1);
+	ews_backend->priv->oauth2_client_id_entry = widget;
 
 	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
+		ews_backend->priv->oauth2_override_check, "active",
 		widget, "sensitive",
 		G_BINDING_SYNC_CREATE);
 
@@ -369,25 +350,108 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 		_("There is not set any default application ID"),
 		g_strdup_printf (_("Default application ID is “%s”"), OFFICE365_CLIENT_ID));
 
-	widget = gtk_label_new_with_mnemonic (_("_Redirect URI:"));
+	/* Translators: 'Tenant ID' here means a term used by Microsoft to identify a company or organization in an Office 365 world.
+	   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
+	widget = gtk_label_new_with_mnemonic (_("_Tenant ID:"));
 	gtk_widget_set_margin_left (widget, 12);
 	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 0, 3, 1, 1);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 0, 2, 1, 1);
 	label = GTK_LABEL (widget);
 
 	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
+		ews_backend->priv->oauth2_override_check, "active",
 		widget, "sensitive",
 		G_BINDING_SYNC_CREATE);
 
 	widget = gtk_entry_new ();
 	gtk_widget_set_hexpand (widget, TRUE);
 	gtk_label_set_mnemonic_widget (label, widget);
-	gtk_grid_attach (priv->oauth2_settings_grid, widget, 1, 3, 1, 1);
-	priv->oauth2_redirect_uri_entry = widget;
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 1, 2, 1, 1);
+	ews_backend->priv->oauth2_tenant_entry = widget;
 
 	e_binding_bind_property (
-		priv->oauth2_override_check, "active",
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	mail_config_ews_backend_set_oauth2_tooltip (widget, OFFICE365_TENANT,
+		/* Translators: 'Tenant' here means a term used by Microsoft to identify a company or organization in an Office 365 world. Same for 'common', it's a default URL path.
+		   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
+		_("Default tenant ID is “common“"),
+		/* Translators: 'Tenant' here means a term used by Microsoft to identify a company or organization in an Office 365 world.
+		   You probably do not want to translate it. More for example here: https://powerbi.microsoft.com/en-us/blog/what-is-a-tenant/ */
+		g_strdup_printf (_("Default tenant ID is “%s”"), OFFICE365_TENANT));
+
+	container = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, container, 0, 3, 2, 1);
+
+	widget = gtk_expander_new_with_mnemonic (_("_Advanced Settings"));
+	gtk_widget_set_margin_left (widget, 12);
+	gtk_box_pack_start (GTK_BOX (container), widget, FALSE, FALSE, 0);
+	expander = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	markup = g_markup_printf_escaped ("(<a href=\"https://wiki.gnome.org/Apps/Evolution/EWS/OAuth2#Alternative_endpoints\">%s</a>)", _("Help…"));
+	widget = gtk_label_new (markup);
+	gtk_label_set_use_markup (GTK_LABEL (widget), TRUE);
+	gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+	gtk_box_pack_start (GTK_BOX (container), widget, FALSE, FALSE, 0);
+	g_free (markup);
+	advanced_help = widget;
+
+	widget = gtk_label_new_with_mnemonic (_("_Endpoint host:"));
+	gtk_widget_set_margin_left (widget, 12);
+	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 0, 4, 1, 1);
+	label = GTK_LABEL (widget);
+	endpoint_host_label = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	widget = gtk_entry_new ();
+	gtk_widget_set_hexpand (widget, TRUE);
+	gtk_label_set_mnemonic_widget (label, widget);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 1, 4, 1, 1);
+	ews_backend->priv->oauth2_endpoint_host_entry = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	markup = g_strdup_printf (_("Default endpoint host is “%s”"), "login.microsoftonline.com");
+	mail_config_ews_backend_set_oauth2_tooltip (widget, OFFICE365_ENDPOINT_HOST,
+		markup,
+		g_strdup_printf (_("Default endpoint host is “%s”"), OFFICE365_ENDPOINT_HOST));
+	g_free (markup);
+
+	widget = gtk_label_new_with_mnemonic (_("Red_irect URI:"));
+	gtk_widget_set_margin_left (widget, 12);
+	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 0, 5, 1, 1);
+	label = GTK_LABEL (widget);
+	redirect_uri_label = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	widget = gtk_entry_new ();
+	gtk_widget_set_hexpand (widget, TRUE);
+	gtk_label_set_mnemonic_widget (label, widget);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 1, 5, 1, 1);
+	ews_backend->priv->oauth2_redirect_uri_entry = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
 		widget, "sensitive",
 		G_BINDING_SYNC_CREATE);
 
@@ -397,30 +461,105 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 		g_strdup_printf (_("Default redirect URI is “%s”"), OFFICE365_REDIRECT_URI));
 	g_free (markup);
 
-	gtk_widget_show_all (GTK_WIDGET (priv->oauth2_settings_grid));
+	widget = gtk_label_new_with_mnemonic (_("Re_source URI:"));
+	gtk_widget_set_margin_left (widget, 12);
+	gtk_misc_set_alignment (GTK_MISC (widget), 1.0, 0.5);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 0, 6, 1, 1);
+	label = GTK_LABEL (widget);
+	resource_uri_label = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	widget = gtk_entry_new ();
+	gtk_widget_set_hexpand (widget, TRUE);
+	gtk_label_set_mnemonic_widget (label, widget);
+	gtk_grid_attach (ews_backend->priv->oauth2_settings_grid, widget, 1, 6, 1, 1);
+	ews_backend->priv->oauth2_resource_uri_entry = widget;
+
+	e_binding_bind_property (
+		ews_backend->priv->oauth2_override_check, "active",
+		widget, "sensitive",
+		G_BINDING_SYNC_CREATE);
+
+	markup = g_strdup_printf (_("Default resource URI is derived from the Host URL, or it can eventually fall back to “%s”."), OFFICE365_FALLBACK_RESOURCE_URI);
+	mail_config_ews_backend_set_oauth2_tooltip (widget, NULL,
+		markup,
+		NULL);
+	g_free (markup);
+
+	gtk_widget_show_all (GTK_WIDGET (ews_backend->priv->oauth2_settings_grid));
+
+	camel_ews_settings_lock (ews_settings);
+
+	gtk_expander_set_expanded (GTK_EXPANDER (expander),
+		(e_util_strcmp0 (camel_ews_settings_get_oauth2_endpoint_host (ews_settings), NULL) != 0 &&
+		 e_util_strcmp0 (camel_ews_settings_get_oauth2_endpoint_host (ews_settings), OFFICE365_ENDPOINT_HOST) != 0) ||
+		(e_util_strcmp0 (camel_ews_settings_get_oauth2_redirect_uri (ews_settings), NULL) != 0 &&
+		 e_util_strcmp0 (camel_ews_settings_get_oauth2_redirect_uri (ews_settings), OFFICE365_REDIRECT_URI) != 0) ||
+		e_util_strcmp0 (camel_ews_settings_get_oauth2_resource_uri (ews_settings), NULL) != 0);
+
+	camel_ews_settings_unlock (ews_settings);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		advanced_help, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		endpoint_host_label, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		ews_backend->priv->oauth2_endpoint_host_entry, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		redirect_uri_label, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		ews_backend->priv->oauth2_redirect_uri_entry, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		resource_uri_label, "visible",
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_property (
+		expander, "expanded",
+		ews_backend->priv->oauth2_resource_uri_entry, "visible",
+		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_property_full (
-		priv->auth_check, "active-mechanism",
-		priv->oauth2_settings_grid, "visible",
+		ews_backend->priv->auth_check, "active-mechanism",
+		ews_backend->priv->oauth2_settings_grid, "visible",
 		G_BINDING_SYNC_CREATE,
 		mail_config_ews_backend_auth_mech_is_oauth2,
 		NULL, NULL, NULL);
 
 	e_binding_bind_object_text_property (
 		settings, "user",
-		priv->user_entry, "text",
+		ews_backend->priv->user_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_object_text_property (
 		settings, "hosturl",
-		priv->host_entry, "text",
+		ews_backend->priv->host_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_object_text_property (
 		settings, "oaburl",
-		priv->oab_entry, "text",
+		ews_backend->priv->oab_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
@@ -429,30 +568,42 @@ mail_config_ews_backend_insert_widgets (EMailConfigServiceBackend *backend,
 	 * a simple property binding would. */
 	e_binding_bind_property (
 		settings, "auth-mechanism",
-		priv->auth_check, "active-mechanism",
+		ews_backend->priv->auth_check, "active-mechanism",
 		G_BINDING_BIDIRECTIONAL);
 
 	e_binding_bind_property (
 		settings, "override-oauth2",
-		priv->oauth2_override_check, "active",
+		ews_backend->priv->oauth2_override_check, "active",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_object_text_property (
 		settings, "oauth2-tenant",
-		priv->oauth2_tenant_entry, "text",
+		ews_backend->priv->oauth2_tenant_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_object_text_property (
 		settings, "oauth2-client-id",
-		priv->oauth2_client_id_entry, "text",
+		ews_backend->priv->oauth2_client_id_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
 	e_binding_bind_object_text_property (
-		settings, "oauth2-redirect_uri",
-		priv->oauth2_redirect_uri_entry, "text",
+		settings, "oauth2-redirect-uri",
+		ews_backend->priv->oauth2_redirect_uri_entry, "text",
+		G_BINDING_BIDIRECTIONAL |
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_object_text_property (
+		settings, "oauth2-resource-uri",
+		ews_backend->priv->oauth2_resource_uri_entry, "text",
+		G_BINDING_BIDIRECTIONAL |
+		G_BINDING_SYNC_CREATE);
+
+	e_binding_bind_object_text_property (
+		settings, "oauth2-endpoint-host",
+		ews_backend->priv->oauth2_endpoint_host_entry, "text",
 		G_BINDING_BIDIRECTIONAL |
 		G_BINDING_SYNC_CREATE);
 
@@ -536,7 +687,7 @@ static gboolean
 mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 {
 	EMailConfigServicePage *page;
-	EMailConfigEwsBackendPrivate *priv;
+	EMailConfigEwsBackend *ews_backend;
 	CamelSettings *settings;
 	CamelEwsSettings *ews_settings;
 	CamelNetworkSettings *network_settings;
@@ -544,7 +695,7 @@ mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 	const gchar *user;
 	gboolean correct, complete = TRUE;
 
-	priv = E_MAIL_CONFIG_EWS_BACKEND_GET_PRIVATE (backend);
+	ews_backend = E_MAIL_CONFIG_EWS_BACKEND (backend);
 	page = e_mail_config_service_backend_get_page (backend);
 
 	/* This backend serves double duty.  One instance holds the
@@ -560,6 +711,9 @@ mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 	settings = e_mail_config_service_backend_get_settings (backend);
 
 	ews_settings = CAMEL_EWS_SETTINGS (settings);
+
+	camel_ews_settings_lock (ews_settings);
+
 	hosturl = camel_ews_settings_get_hosturl (ews_settings);
 	oaburl = camel_ews_settings_get_oaburl (ews_settings);
 
@@ -575,13 +729,13 @@ mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 		suri = soup_uri_new (hosturl);
 		if (suri) {
 			soup_uri_free (suri);
-			e_util_set_entry_issue_hint (priv->host_entry, NULL);
+			e_util_set_entry_issue_hint (ews_backend->priv->host_entry, NULL);
 		} else {
-			e_util_set_entry_issue_hint (priv->host_entry, _("Host URL is not valid"));
+			e_util_set_entry_issue_hint (ews_backend->priv->host_entry, _("Host URL is not valid"));
 			complete = FALSE;
 		}
 	} else {
-		e_util_set_entry_issue_hint (priv->host_entry, _("Host URL cannot be empty"));
+		e_util_set_entry_issue_hint (ews_backend->priv->host_entry, _("Host URL cannot be empty"));
 	}
 
 	if (oaburl && *oaburl) {
@@ -590,19 +744,19 @@ mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 		suri = soup_uri_new (oaburl);
 		if (suri) {
 			soup_uri_free (suri);
-			e_util_set_entry_issue_hint (priv->oab_entry, NULL);
+			e_util_set_entry_issue_hint (ews_backend->priv->oab_entry, NULL);
 		} else {
-			e_util_set_entry_issue_hint (priv->oab_entry, _("OAB URL is not valid"));
+			e_util_set_entry_issue_hint (ews_backend->priv->oab_entry, _("OAB URL is not valid"));
 			complete = FALSE;
 		}
 	} else {
-		e_util_set_entry_issue_hint (priv->oab_entry, NULL);
+		e_util_set_entry_issue_hint (ews_backend->priv->oab_entry, NULL);
 	}
 
 	correct = user != NULL && *user != '\0';
 	complete = complete && correct;
 
-	e_util_set_entry_issue_hint (priv->user_entry, correct ? NULL : _("User name cannot be empty"));
+	e_util_set_entry_issue_hint (ews_backend->priv->user_entry, correct ? NULL : _("User name cannot be empty"));
 
 	if (correct && camel_ews_settings_get_auth_mechanism (ews_settings) == EWS_AUTH_TYPE_OAUTH2) {
 		const gchar *client_id;
@@ -613,12 +767,14 @@ mail_config_ews_backend_check_complete (EMailConfigServiceBackend *backend)
 			client_id = OFFICE365_CLIENT_ID;
 		}
 
-		correct = client_id && *client_id;
+		correct = e_util_strcmp0 (client_id, NULL) != 0;
 		complete = complete && correct;
 
-		e_util_set_entry_issue_hint (priv->oauth2_client_id_entry, correct ? NULL : _("Application ID cannot be empty"));
+		e_util_set_entry_issue_hint (ews_backend->priv->oauth2_client_id_entry, correct ? NULL : _("Application ID cannot be empty"));
 	}
 
+	camel_ews_settings_unlock (ews_settings);
+
 	return complete;
 }
 
@@ -656,9 +812,6 @@ e_mail_config_ews_backend_class_init (EMailConfigEwsBackendClass *class)
 {
 	EMailConfigServiceBackendClass *backend_class;
 
-	g_type_class_add_private (
-		class, sizeof (EMailConfigEwsBackendPrivate));
-
 	backend_class = E_MAIL_CONFIG_SERVICE_BACKEND_CLASS (class);
 	backend_class->backend_name = "ews";
 	backend_class->new_collection = mail_config_ews_backend_new_collection;
@@ -677,7 +830,7 @@ e_mail_config_ews_backend_class_finalize (EMailConfigEwsBackendClass *class)
 static void
 e_mail_config_ews_backend_init (EMailConfigEwsBackend *backend)
 {
-	backend->priv = E_MAIL_CONFIG_EWS_BACKEND_GET_PRIVATE (backend);
+	backend->priv = e_mail_config_ews_backend_get_instance_private (backend);
 }
 
 void
